from .training import train, regLoss
#from .training import multitaskMSE, multitaskTrain, multitaskLoss
from .training import test
from .training import LangevinOpt
from .networks import FCNet
from .networks import ConvNet