from .theory_1HL_FC_single_output import FC_1HL, FC_1HL_corrected, FC_1HL_nonodd
from .theory_1HL_FC_single_output_cuda import FC_1HL_cuda
from .theory_1HL_FC_multiple_output import FC_1HL_multiclass
from .theory_deep_FC import FC_deep
from .theory_1HL_CONV import CONV_1HL, validation_CONV_1HL
#from .theory_1HL_extras import FC_1HL_spectral, FC_1HL_multitask